// Copyright (C) 2001-2003 Vivien Mallet
//
// This file is part of Seldon library.
// Seldon library provides matrices and vectors structures for
// linear algebra.
// 
// Seldon is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Seldon is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License (file "license") for more details.
//
// For more information, please see the Seldon home page:
//     http://spacetown.free.fr/lib/seldon/

#ifndef FILE_FUNCTIONS_MATVECT_CXX

/*
  Functions defined in this file:

  alpha.A*U -> C
  Mlt(alpha, A, U, B)

  alpha.A*U -> U
  Mlt(alpha, A, U)

  alpha.A*U + V -> V
  MltAdd(alpha, A, U, V)
*/

namespace Seldon
{


  /////////
  // Mlt //


  template <class T0,
	    class T1, class Prop1, class Storage1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const Matrix<T1, Prop1, Storage1, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    int ma=A.GetM();
    int na=A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if (na!=B.GetM()) throw WrongDim("Mlt(Scalar, const Matrix, const Vector, Vector)", "Unable to multiply matrix and vector.");
#endif

    C.Reallocate(ma);

    T3 zero(0);
    T3 temp;
    T3 alpha_(alpha);

    for (int i=0; i<ma; i++)
      {
	temp=zero;
	for (int j=0; j<na; j++)
	  temp+=A(i,j)*B(j);
	C(i)=alpha_*temp;
      }
  }


  // Mlt //
  /////////



  /////////
  // Mlt //


  // For sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const Matrix<T1, Prop1, RowSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    int ma=A.GetM();
    int na=A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if (na!=B.GetM()) throw WrongDim("Mlt(Scalar, const Matrix, const Vector, Vector)", "Unable to multiply matrix and vector.");
#endif

    C.Reallocate(ma);
    C.Zero();

    T3 zero(0);
    T3 temp;
    T3 alpha_(alpha);

    int* ptr=A.GetPtr();
    int* ind=A.GetInd();
    typename Matrix<T1, Prop1, RowSparse, Allocator1>::pointer data=A.GetData();

    for (int i=0; i<ma; i++)
      {
	temp=zero;
	for (int j=ptr[i]; j<ptr[i+1]; j++)
	  temp+=data[j]*B(ind[j]);
	C(i)=alpha_*temp;
      }
  }


  // Mlt //
  /////////



  /////////
  // Mlt //


  // For complex sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const Matrix<T1, Prop1, RowComplexSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    int i, j;

    int ma = A.GetM();
    int na = A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if (na!=B.GetM()) throw WrongDim("Mlt(Scalar, const Matrix, const Vector, Vector)", "Unable to multiply matrix and vector.");
#endif

    C.Reallocate(ma);
    C.Zero();

    T3 zero(0);
    T3 temp;

    int* real_ptr = A.GetRealPtr();
    int* imag_ptr = A.GetImagPtr();
    int* real_ind = A.GetRealInd();
    int* imag_ind = A.GetImagInd();
    typename Matrix<T1, Prop1, RowComplexSparse, Allocator1>::pointer
      real_data = A.GetRealData();
    typename Matrix<T1, Prop1, RowComplexSparse, Allocator1>::pointer
      imag_data = A.GetImagData();

    for (i=0; i<ma; i++)
      {
	temp = zero;
	for (j=real_ptr[i]; j<real_ptr[i+1]; j++)
	  temp += real_data[j] * B(real_ind[j]);
	C(i) = temp;
      }

    for (i=0; i<ma; i++)
      {
	temp = zero;
	for (j=imag_ptr[i]; j<imag_ptr[i+1]; j++)
	  temp += complex<T1>(T1(0), imag_data[j]) * B(imag_ind[j]);
	C(i) += temp;
      }

    Mlt(alpha, C);
  }


  // Mlt //
  /////////



  /////////
  // Mlt //


  // For symmetric sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const Matrix<T1, Prop1, RowSymSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    int ma = A.GetM();
    int na = A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if (na!=B.GetM())
      throw WrongDim("Mlt(Scalar, const Matrix, const Vector, Vector)",
		     "Unable to multiply matrix and vector.");
#endif

    C.Reallocate(ma);

    int i, j;
    T3 zero(0);
    T3 temp;
    T3 alpha_(alpha);

    int* ptr = A.GetPtr();
    int* ind = A.GetInd();
    typename Matrix<T1, Prop1, RowSymSparse, Allocator1>::pointer data = A.GetData();

    for (i=0; i<ma; i++)
      {
	temp = zero;
	for (j=ptr[i]; j<ptr[i+1]; j++)
	  temp += data[j] * B(ind[j]);
	C(i) = temp;
      }
    for (i=0; i<ma-1; i++)
      for (j=ptr[i]; j<ptr[i+1]; j++)
	if (ind[j]!=i)
	  C(ind[j]) += data[j] * B(i);
    Mlt(alpha_, C);
  }


  // Mlt //
  /////////



  /////////
  // Mlt //


  // For symmetric complex sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const Matrix<T1, Prop1, RowSymComplexSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    int ma = A.GetM();

#ifdef SELDON_CHECK_BOUNDARIES
    if (A.GetN()!=B.GetM())
      throw WrongDim("Mlt(Scalar, const Matrix, const Vector, Vector)",
		     "Unable to multiply matrix and vector.");
#endif

    C.Reallocate(ma);

    int i, j;
    T3 zero(0);
    T3 temp;
    T3 alpha_(alpha);

    int* real_ptr = A.GetRealPtr();
    int* imag_ptr = A.GetImagPtr();
    int* real_ind = A.GetRealInd();
    int* imag_ind = A.GetImagInd();
    typename Matrix<T1, Prop1, RowSymComplexSparse, Allocator1>::pointer
      real_data = A.GetRealData();
    typename Matrix<T1, Prop1, RowSymComplexSparse, Allocator1>::pointer
      imag_data = A.GetImagData();

    for (i=0; i<ma; i++)
      {
	temp = zero;
	for (j=real_ptr[i]; j<real_ptr[i+1]; j++)
	  temp += real_data[j] * B(real_ind[j]);
	C(i) = temp;
      }
    for (i=0; i<ma-1; i++)
      for (j=real_ptr[i]; j<real_ptr[i+1]; j++)
	if (real_ind[j]!=i)
	  C(real_ind[j]) += real_data[j] * B(i);

    for (i=0; i<ma; i++)
      {
	temp = zero;
	for (j=imag_ptr[i]; j<imag_ptr[i+1]; j++)
	  temp += complex<T1>(T1(0), imag_data[j]) * B(real_ind[j]);
	C(i) += temp;
      }
    for (i=0; i<ma-1; i++)
      for (j=imag_ptr[i]; j<imag_ptr[i+1]; j++)
	if (imag_ind[j]!=i)
	  C(imag_ind[j]) += complex<T1>(T1(0), imag_data[j]) * B(i);

    Mlt(alpha_, C);
  }


  // Mlt //
  /////////



  /////////
  // Mlt //


  template <class T0,
	    class T1, class Prop1, class Storage1, class Allocator1,
	    class T2, class Storage2, class Allocator2>
  void Mlt(const T0 alpha,
	   const Matrix<T1, Prop1, Storage1, Allocator1>& A,
	   Vector<T2, Storage2, Allocator2>& B)
  {
    int ma=A.GetM();
    int na=A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if ((na!=B.GetM()) || (ma!=B.GetM())) throw WrongDim("Mlt(Scalar, const Matrix, const Vector, Vector)", "Unable to multiply matrix and vector.");
#endif

    Vector<T2, Storage2, Allocator2> C(B);

    T2 zero(0);
    T2 temp;
    T2 alpha_(alpha);

    for (int i=0; i<ma; i++)
      {
	temp=zero;
	for (int j=0; j<na; j++)
	  temp+=A(i,j)*C(j);
	B(i)=alpha_*temp;
      }
  }


  // Mlt //
  /////////



  /////////
  // Mlt //


  // For sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const class_SeldonNoTrans& Trans,
	   const Matrix<T1, Prop1, RowSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    Mlt(alpha, A, B, C);
  }


  // For sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const class_SeldonTrans& Trans,
	   const Matrix<T1, Prop1, RowSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    int i, j;

    int ma=A.GetM();
    int na=A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if (ma!=B.GetM()) throw WrongDim("MltTrans(Scalar, const Matrix, const Vector, Vector)", "Unable to multiply matrix and vector.");
#endif

    C.Reallocate(na);
    C.Zero();

    T3 alpha_(alpha);

    int* ptr=A.GetPtr();
    int* ind=A.GetInd();
    typename Matrix<T1, Prop1, RowSparse, Allocator1>::pointer data=A.GetData();

    for (i=0; i<ma; i++)
      for (j=ptr[i]; j<ptr[i+1]; j++)
	C(ind[j]) += data[j] * B(i);

    Mlt(alpha_, C);
  }


  // For sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const class_SeldonConjTrans& Trans,
	   const Matrix<complex<T1>, Prop1, RowSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    int i, j;

    int ma=A.GetM();
    int na=A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if (ma!=B.GetM())
      throw WrongDim("Mlt(Scalar, class_SeldonConjTrans, const Matrix, const Vector, Vector)",
		     "Unable to multiply matrix and vector.");
#endif

    C.Reallocate(na);
    C.Zero();

    T3 alpha_(alpha);

    int* ptr=A.GetPtr();
    int* ind=A.GetInd();
    typename Matrix<complex<T1>, Prop1, RowSparse, Allocator1>::pointer
      data=A.GetData();

    for (i=0; i<ma; i++)
      for (j=ptr[i]; j<ptr[i+1]; j++)
	C(ind[j]) += conj(data[j]) * B(i);

    Mlt(alpha_, C);
  }


  // Mlt //
  /////////



  /////////
  // Mlt //


  // For complex sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const class_SeldonNoTrans& Trans,
	   const Matrix<T1, Prop1, RowComplexSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    Mlt(alpha, A, B, C);
  }


  // For complex sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const class_SeldonTrans& Trans,
	   const Matrix<T1, Prop1, RowComplexSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    int i, j;

    int ma = A.GetM();
    int na = A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if (na!=B.GetM()) throw WrongDim("Mlt(Scalar, const Matrix, const Vector, Vector)", "Unable to multiply matrix and vector.");
#endif

    C.Reallocate(ma);
    C.Zero();

    int* real_ptr = A.GetRealPtr();
    int* imag_ptr = A.GetImagPtr();
    int* real_ind = A.GetRealInd();
    int* imag_ind = A.GetImagInd();
    typename Matrix<T1, Prop1, RowComplexSparse, Allocator1>::pointer
      real_data = A.GetRealData();
    typename Matrix<T1, Prop1, RowComplexSparse, Allocator1>::pointer
      imag_data = A.GetImagData();

    for (i=0; i<ma; i++)
      for (j=real_ptr[i]; j<real_ptr[i+1]; j++)
	C(real_ind[j]) += real_data[j] * B(i);

    for (i=0; i<ma; i++)
      for (j=imag_ptr[i]; j<imag_ptr[i+1]; j++)
	C(imag_ind[j]) += complex<T1>(T1(0), imag_data[j]) * B(i);

    Mlt(alpha, C);
  }


  // For complex sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const class_SeldonConjTrans& Trans,
	   const Matrix<T1, Prop1, RowComplexSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    int i, j;

    int ma = A.GetM();
    int na = A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if (na!=B.GetM()) throw WrongDim("Mlt(Scalar, const Matrix, const Vector, Vector)", "Unable to multiply matrix and vector.");
#endif

    C.Reallocate(ma);
    C.Zero();

    int* real_ptr = A.GetRealPtr();
    int* imag_ptr = A.GetImagPtr();
    int* real_ind = A.GetRealInd();
    int* imag_ind = A.GetImagInd();
    typename Matrix<T1, Prop1, RowComplexSparse, Allocator1>::pointer
      real_data = A.GetRealData();
    typename Matrix<T1, Prop1, RowComplexSparse, Allocator1>::pointer
      imag_data = A.GetImagData();

    for (i=0; i<ma; i++)
      for (j=real_ptr[i]; j<real_ptr[i+1]; j++)
	C(real_ind[j]) += real_data[j] * B(i);

    for (i=0; i<ma; i++)
      for (j=imag_ptr[i]; j<imag_ptr[i+1]; j++)
	C(imag_ind[j]) += complex<T1>(T1(0), - imag_data[j]) * B(i);

    Mlt(alpha, C);
  }


  // Mlt //
  /////////



  /////////
  // Mlt //


  // For symmetric sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const class_SeldonNoTrans& Trans,
	   const Matrix<T1, Prop1, RowSymSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    Mlt(alpha, A, B, C);
  }


  // For symmetric sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const class_SeldonTrans& Trans,
	   const Matrix<T1, Prop1, RowSymSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    Mlt(alpha, A, B, C);
  }


  // For symmetric sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const class_SeldonConjTrans& Trans,
	   const Matrix<T1, Prop1, RowSymSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    int ma = A.GetM();
    int na = A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if (na!=B.GetM())
      throw WrongDim("Mlt(Scalar, class_SeldonConjTrans, const Matrix, const Vector, Vector)",
		     "Unable to multiply matrix and vector.");
#endif

    C.Reallocate(ma);

    int i, j;
    T3 zero(0);
    T3 temp;
    T3 alpha_(alpha);

    int* ptr = A.GetPtr();
    int* ind = A.GetInd();
    typename Matrix<complex<T1>, Prop1, RowSymSparse, Allocator1>::pointer
      data = A.GetData();

    for (i=0; i<ma; i++)
      {
	temp = zero;
	for (j=ptr[i]; j<ptr[i+1]; j++)
	  temp += conj(data[j]) * B(ind[j]);
	C(i) = temp;
      }
    for (i=0; i<ma-1; i++)
      for (j=ptr[i]; j<ptr[i+1]; j++)
	if (ind[j]!=i)
	  C(ind[j]) += conj(data[j]) * B(i);
    Mlt(alpha_, C);
  }


  // Mlt //
  /////////



  /////////
  // Mlt //


  // For symmetric complex sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const class_SeldonNoTrans& Trans,
	   const Matrix<T1, Prop1, RowSymComplexSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    Mlt(alpha, A, B, C);
  }


  // For symmetric complex sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const class_SeldonTrans& Trans,
	   const Matrix<T1, Prop1, RowSymComplexSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    Mlt(alpha, A, B, C);
  }


  // For symmetric complex sparse matrices.
  template <class T0,
	    class T1, class Prop1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const class_SeldonConjTrans& Trans,
	   const Matrix<T1, Prop1, RowSymComplexSparse, Allocator1>& A,
	   const Vector<T2, Storage2, Allocator2>& B,
	   Vector<T3, Storage3, Allocator3>& C)
  {
    int ma = A.GetM();

#ifdef SELDON_CHECK_BOUNDARIES
    if (A.GetN()!=B.GetM())
      throw WrongDim("Mlt(Scalar, class_SeldonConjTrans, const Matrix, const Vector, Vector)",
		     "Unable to multiply matrix and vector.");
#endif

    C.Reallocate(ma);

    int i, j;
    T3 zero(0);
    T3 temp;
    T3 alpha_(alpha);

    int* real_ptr = A.GetRealPtr();
    int* imag_ptr = A.GetImagPtr();
    int* real_ind = A.GetRealInd();
    int* imag_ind = A.GetImagInd();
    typename Matrix<T1, Prop1, RowSymComplexSparse, Allocator1>::pointer
      real_data = A.GetRealData();
    typename Matrix<T1, Prop1, RowSymComplexSparse, Allocator1>::pointer
      imag_data = A.GetImagData();

    for (i=0; i<ma; i++)
      {
	temp = zero;
	for (j=real_ptr[i]; j<real_ptr[i+1]; j++)
	  temp += real_data[j] * B(real_ind[j]);
	C(i) = temp;
      }
    for (i=0; i<ma-1; i++)
      for (j=real_ptr[i]; j<real_ptr[i+1]; j++)
	if (real_ind[j]!=i)
	  C(real_ind[j]) += real_data[j] * B(i);

    for (i=0; i<ma; i++)
      {
	temp = zero;
	for (j=imag_ptr[i]; j<imag_ptr[i+1]; j++)
	  temp += complex<T1>(T1(0), - imag_data[j]) * B(real_ind[j]);
	C(i) += temp;
      }
    for (i=0; i<ma-1; i++)
      for (j=imag_ptr[i]; j<imag_ptr[i+1]; j++)
	if (imag_ind[j]!=i)
	  C(imag_ind[j]) += complex<T1>(T1(0), - imag_data[j]) * B(i);

    Mlt(alpha_, C);
  }


  // Mlt //
  /////////



  ////////////
  // MltAdd //


  template <class T0,
	    class T1, class Prop1, class Storage1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3, class Storage3, class Allocator3>
  void MltAdd(const T0 alpha,
	      const Matrix<T1, Prop1, Storage1, Allocator1>& A,
	      const Vector<T2, Storage2, Allocator2>& B,
	      Vector<T3, Storage3, Allocator3>& C)
  {
    int ma=A.GetM();
    int na=A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if ((na!=B.GetM()) || (na!=C.GetM())) throw WrongDim("Mlt(Scalar, const Matrix, const Vector, Vector)", "Unable to multiply matrix and vector or to add vectors.");
#endif

    T3 zero(0);
    T3 temp;
    T3 alpha_(alpha);

    for (int i=0; i<ma; i++)
      {
	temp=zero;
	for (int j=0; j<na; j++)
	  temp+=A(i,j)*B(j);
	C(i)+=alpha_*temp;
      }
  }


  // MltAdd //
  ////////////



  /////////////////
  // Runge-Kutta //


  // Runge-Kutta (4th order).
  template <class T0, class Prop0, class Storage0, class Storage00, class Allocator0>
  inline void RK4(const T0 delta_t,
		  const Vector<T0, Storage0, Allocator0>& coeff,
		  const Matrix<T0, Prop0, Storage00, Allocator0>& A,
		  Vector<T0, Storage0, Allocator0>& B,
		  Vector<T0, Storage0, Allocator0>& C,
		  Vector<T0, Storage0, Allocator0>& D)
  {
    int j, k;

    int lgth=D.GetLength();
    int size=lgth*sizeof(T0);
    void* datab=reinterpret_cast<void*>(B.GetData());
    void* datac=reinterpret_cast<void*>(C.GetData());

    T0 zero(0);
    T0 temp;

    for (int i=0; i<4; i++)
      {
	for (j=0; j<lgth; j++)
	  {
	    temp=zero;
	    for (k=0; k<lgth; k++)
	      temp+=A(j,k)*B(k);
	    C(j)=D(j)+delta_t*coeff(i)*temp;
	  }
	memcpy(datab, datac, size);
      }

    memcpy(reinterpret_cast<void*>(D.GetData()), datab, size);
  }


  // Rugnge-Kutta //
  //////////////////



  ///////////
  // Gauss //


  // Solve X = A*Y with Gauss method.
  // Warning: A is modified. The results are stored in X.
  template <class T0, class Prop0, class Storage0, class Allocator0,
	    class T1, class Storage1, class Allocator1>
  inline void Gauss(Matrix<T0, Prop0, Storage0, Allocator0>& A,
		    Vector<T1, Storage1, Allocator1>& X)
  {
    int i, j, k;
    T1 r, S;

    int ma=A.GetM();
    int na=A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if (na!=ma) throw WrongDim("Gauss(const Matrix, const Vector, Vector)", "Matrix must be squared.");
    if (ma!=X.GetLength()) throw WrongDim("Gauss(const Matrix, const Vector, Vector)", "Matrix and vector dimensions are incompatible.");
#endif

    for (k=0; k<ma-1; k++)
      for (i=k+1; i<ma; i++)
	{
	  r=A(i,k)/A(k,k);
	  for (j=k+1; j<ma; j++)
	    A(i,j)-=r*A(k,j);
	  X(i)-=r*=X(k);
	}

    X(ma-1)=X(ma-1)/A(ma-1,ma-1);
    for (k=ma-2; k>-1; k--)
      {
	S=X(k);
	for (j=k+1; j<ma; j++)
	  S-=A(k,j)*X(j);
	X(k)=S/A(k,k);
      }
  }


  // Gauss //
  ///////////



  ////////////////////
  // Gauss - Seidel //


  // Solve X = A*Y with Gauss-Seidel method.
  template <class T0, class Prop0, class Storage0, class Allocator0,
	    class T1, class Storage1, class Allocator1,
	    class T2, class Storage2, class Allocator2>
  inline void GaussSeidel(const Matrix<T0, Prop0, Storage0, Allocator0>& A,
			  const Vector<T1, Storage1, Allocator1>& X,
			  Vector<T2, Storage2, Allocator2>& Y,
			  int iter)
  {
    int i, j, k;
    T1 temp;

    int ma=A.GetM();
    int na=A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if (na!=ma) throw WrongDim("GaussSeidel(const Matrix, const Vector, Vector)", "Matrix must be squared.");
    if ((ma!=X.GetLength()) || (ma!=Y.GetLength())) throw WrongDim("GaussSeidel(const Matrix, const Vector, Vector)", "Matrix and vector dimensions are incompatible.");
#endif
    
    for (i=0; i<iter; i++)
      for (j=0; j<na; j++)
	{
	  temp=0;
	  for (k=0; k<j; k++)
	    temp-=A(j,k)*Y(k);
	  for (k=j+1; k<na; k++)
	    temp-=A(j,k)*Y(k);
	  Y(j)=(X(j)+temp)/A(j,j);
	}
  }


  // Gauss-Seidel //
  //////////////////



  ///////////////////
  // S.O.R. method //


  // Solve X = A*Y with S.O.R. method.
  template <class T0, class Prop0, class Storage0, class Allocator0,
	    class T1, class Storage1, class Allocator1,
	    class T2, class Storage2, class Allocator2,
	    class T3>
  inline void SOR(const Matrix<T0, Prop0, Storage0, Allocator0>& A,
		  const Vector<T1, Storage1, Allocator1>& X,
		  Vector<T2, Storage2, Allocator2>& Y,
		  T3 omega,
		  int iter)
  {
    int i, j, k;
    T1 temp;

    int ma=A.GetM();
    int na=A.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if (na!=ma) throw WrongDim("GaussSeidel(const Matrix, const Vector, Vector)", "Matrix must be squared.");
    if ((ma!=X.GetLength()) || (ma!=Y.GetLength())) throw WrongDim("GaussSeidel(const Matrix, const Vector, Vector)", "Matrix and vector dimensions are incompatible.");
#endif
    
    for (i=0; i<iter; i++)
      for (j=0; j<na; j++)
	{
	  temp=0;
	  for (k=0; k<j; k++)
	    temp-=A(j,k)*Y(k);
	  for (k=j+1; k<na; k++)
	    temp-=A(j,k)*Y(k);
	  Y(j)=(1-omega)*Y(j)+omega*(X(j)+temp)/A(j,j);
	}
  }


  // Gauss-Seidel //
  //////////////////



  /////////////
  // SolveLU //


  // Solves A.X=V where A has been decomposed in a LU form. V is overwritten (V <- X).
  template <class T0, class Prop0, class Storage0, class Allocator0,
	    class T1, class Storage1, class Allocator1>
  void SolveLU(const Matrix<T0, Prop0, Storage0, Allocator0>& A,
	       Vector<T1, Storage1, Allocator1>& V)
  {
    int i, q, k;
    T0 temp;

    int ma=A.GetM();

#ifdef SELDON_CHECK_BOUNDARIES
    int na=A.GetN();
    if (na!=ma) throw WrongDim("SolveLU(const Matrix, Vector)", "Matrix must be squared.");
    if (ma!=V.GetM()) throw WrongDim("SolveLu(const Matrix, Vector)", "Matrix and vector dimensions are incompatible.");
#endif

    // Forward substitution.
    for (i=0; i<ma; i++)
      {
	temp=0;
	for (k=0; k<i; k++)
	  temp+=A(i,k)*V(k);
	V(i)=(V(i)-temp)/A(i,i);
      }
    // Back substitution.
    for (i=ma-2; i>-1; i--)
      {
	temp=0;
	for (k=i+1; k<ma; k++)
	  temp+=A(i,k)*V(k);
	V(i)-=temp;
      }

  }


  // SolveLU //
  /////////////



}  // namespace Seldon

#define SELDON_FUNCTIONS_MATVECT_CXX
#endif
