// Copyright (C) 2001-2003 Vivien Mallet
//
// This file is part of Seldon library.
// Seldon library provides matrices and vectors structures for
// linear algebra.
// 
// Seldon is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Seldon is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License (file "license") for more details.
//
// For more information, please see the Seldon home page:
//     http://spacetown.free.fr/lib/seldon/

#ifndef FILE_FUNCTIONS_MATRIX_CXX

/*
  Function defined in this file:

  alpha.A*B -> C
  Mlt(alpha, A, B, C)

  alpha.A -> A
  Mlt(alpha, available

  alpha.A + beta.C -> C
  Add(alpha, A, beta, C)

  alpha.A + beta.B -> C
  Add(alpha, A, beta, B, C)
*/

namespace Seldon
{


  /////////
  // Mlt //


  template <class T0,
	    class T1, class Prop1, class Storage1, class Allocator1,
	    class T2, class Prop2, class Storage2, class Allocator2,
	    class T3, class Prop3, class Storage3, class Allocator3>
  void Mlt(const T0 alpha,
	   const Matrix<T1, Prop1, Storage1, Allocator1>& A,
	   const Matrix<T2, Prop2, Storage2, Allocator2>& B,
	   Matrix<T3, Prop3, Storage3, Allocator3>& C)  throw(WrongDim, NoMemory)
  {
    int ma=A.GetM();
    int na=A.GetN();
    int nb=B.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    if (na!=B.GetM()) throw WrongDim("Mlt(Scalar, const Matrix, const Matrix, Matrix)", "Unable to multiply matrices.");
#endif

    C.Reallocate(ma, nb);

    T3 zero(0);
    T3 temp;
    T3 alpha_(alpha);

    for (int i=0; i<Storage3::GetFirst(ma, nb); i++)
      for (int j=0; j<Storage3::GetSecond(ma, nb); j++)
	{
	  temp=zero;
	  for (int k=0; k<na; k++)
	    temp+=A(Storage3::GetFirst(i,j), k)*B(k, Storage3::GetSecond(i,j));
	  C(Storage3::GetFirst(i,j), Storage3::GetSecond(i,j))=alpha_*temp;
	}
  }  


  // Mlt //
  /////////



  /////////
  // Mlt //


  template <class T0,
	    class T1, class Prop1, class Storage1, class Allocator1>
  void Mlt(const T0 alpha,
	   Matrix<T1, Prop1, Storage1, Allocator1>& A)  throw()
  {
    T1 alpha_=alpha;

    typename Matrix<T1, Prop1, Storage1, Allocator1>::pointer data=A.GetData();

    for (int i=0; i<A.GetDataSize(); i++)
      data[i]=alpha_*data[i];
  }  


  // Mlt //
  /////////



  /////////
  // Add //


  template <class T0,
	    class T1, class Prop1, class Storage1, class Allocator1,
	    class T2,
	    class T3, class Prop3, class Storage3, class Allocator3>
  void Add(const T0 alpha,
	   const Matrix<T1, Prop1, Storage1, Allocator1>& A,
	   const T2 beta,
	   Matrix<T3, Prop3, Storage3, Allocator3>& C)  throw(WrongDim, NoMemory)
  {
    if ((alpha!=T0(0)) && (beta!=T2(0)))
      {
	T1 alpha_=alpha;
	T3 beta_=beta;

	int ma=A.GetM();
	int na=A.GetN();
	
#ifdef SELDON_CHECK_BOUNDARIES
	if ((ma!=C.GetM()) || (na!=C.GetN())) throw WrongDim("Add(Scalar, const Matrix, Scalar, Matrix)", "Unable to add matrices.");
#endif

	for (int i=0; i<Storage3::GetFirst(ma, na); i++)
	  for (int j=0; j<Storage3::GetSecond(ma, na); j++)
	    C(Storage3::GetFirst(i,j), Storage3::GetSecond(i,j))=beta_*C(Storage3::GetFirst(i,j), Storage3::GetSecond(i,j))+alpha_*A(Storage3::GetFirst(i,j), Storage3::GetSecond(i,j));
      }
    else if ((alpha==T0(0)) && (beta==T2(0)))
      {
	C.Zero();
      }
    else if (beta!=T2(0))
      {
	T3 beta_=beta;
	
	typename Matrix<T3, Prop3, Storage3, Allocator3>::pointer data=C.GetData();

	for(int i=0; i<C.GetDataSize(); i++)
	  data[i]=beta_*data[i];
      }
    else
      {
	int ma=A.GetM();
	int na=A.GetN();
	
	C.Reallocate(ma, na);

	T1 alpha_=alpha;
	T3 beta_=beta;
	
	for (int i=0; i<Storage3::GetFirst(ma, na); i++)
	  for (int j=0; j<Storage3::GetSecond(ma, na); j++)
	    C(Storage3::GetFirst(i,j), Storage3::GetSecond(i,j))=alpha_*A(Storage3::GetFirst(i,j), Storage3::GetSecond(i,j));
      }
  }


  // Add //
  /////////



  /////////
  // Add //


  template <class T0,
	    class T1, class Prop1, class Storage1, class Allocator1,
	    class T2,
	    class T3, class Prop3, class Storage3, class Allocator3,
	    class T4, class Prop4, class Storage4, class Allocator4>
  void Add(const T0 alpha,
	   const Matrix<T1, Prop1, Storage1, Allocator1>& A,
	   const T2 beta,
	   const Matrix<T3, Prop3, Storage3, Allocator3>& B,
	   Matrix<T4, Prop4, Storage4, Allocator4>& C)  throw(WrongDim, NoMemory)
  {
    if ((alpha!=T0(0)) && (beta!=T2(0)))
      {
	T1 alpha_=alpha;
	T3 beta_=beta;

	int ma=A.GetM();
	int na=A.GetN();
	
#ifdef SELDON_CHECK_BOUNDARIES
	if ((ma!=B.GetM()) || (na!=B.GetN())) throw WrongDim("Add(Scalar, const Matrix, Scalar, const Matrix, Matrix)", "Unable to add matrices.");
#endif

	C.Reallocate(ma, na);

	for (int i=0; i<Storage4::GetFirst(ma, na); i++)
	  for (int j=0; j<Storage4::GetSecond(ma, na); j++)
	    C(Storage4::GetFirst(i,j), Storage4::GetSecond(i,j))=alpha_*A(Storage4::GetFirst(i,j), Storage4::GetSecond(i,j))+beta_*B(Storage4::GetFirst(i,j), Storage4::GetSecond(i,j));
      }
    else if ((alpha==T0(0)) && (beta==T2(0)))
      {
	int ma=A.GetM();
	int na=A.GetN();

	C.Reallocate(ma, na);
	C.Zero();
      }
    else if (beta!=T3(0))
      {
	int mb=B.GetM();
	int nb=B.GetN();

	T3 beta_=beta;
	
	C.Reallocate(mb, nb);

	for (int i=0; i<Storage4::GetFirst(mb, nb); i++)
	  for (int j=0; j<Storage4::GetSecond(mb, nb); j++)
	    C(Storage4::GetFirst(i,j), Storage4::GetSecond(i,j))=beta_*B(Storage4::GetFirst(i,j), Storage4::GetSecond(i,j));
      }
    else
      {
	int ma=A.GetM();
	int na=A.GetN();
	
	T1 alpha_=alpha;

	C.Reallocate(ma, na);

	for (int i=0; i<Storage4::GetFirst(ma, na); i++)
	  for (int j=0; j<Storage4::GetSecond(ma, na); j++)
	    C(Storage4::GetFirst(i,j), Storage4::GetSecond(i,j))=alpha_*A(Storage4::GetFirst(i,j), Storage4::GetSecond(i,j));
      }
  }


  // Add //
  /////////



  ///////////
  // GetLU //


  // Returns the LU decomposition of A=LU (in A) where L diagonal elements are set to unit value.
  template <class T0, class Prop0, class Storage0, class Allocator0>
  void GetLU(Matrix<T0, Prop0, Storage0, Allocator0>& A)  throw(WrongDim)
  {
    int i, p, q, k;
    T0 temp;

    int ma=A.GetM();

#ifdef SELDON_CHECK_BOUNDARIES
    int na=A.GetN();
    if (na!=ma) throw WrongDim("GetLU(Matrix)", "Matrix must be squared.");
#endif

    for (i=0; i<ma; i++)
      {
	for (p=i; p<ma; p++)
	  {
	    temp=0;
	    for (k=0; k<i; k++)
	      temp+=A(p,k)*A(k,i);
	    A(p,i)-=temp;
	  }
	for (q=i+1; q<ma; q++)
	  {
	    temp=0;
	    for (k=0; k<i; k++)
	      temp+=A(i,k)*A(k,q);
	    A(i,q)=(A(i,q)-temp)/A(i,i);
	  }
      }

  }
  
  
  // GetLU //
  ///////////



  /////////////
  // SolveLU //


  // Solves A.X=B where A has been decomposed in a LU form. B is overwritten.
  template <class T0, class Prop0, class Storage0, class Allocator0,
	    class T1, class Prop1, class Storage1, class Allocator1>
  void SolveLU(const Matrix<T0, Prop0, Storage0, Allocator0>& A,
	       Matrix<T1, Prop1, Storage1, Allocator1>& B)  throw(WrongDim)
  {
    int i, q, k;
    T0 temp;

    int ma=A.GetM();
    int nb=B.GetN();

#ifdef SELDON_CHECK_BOUNDARIES
    int na=A.GetN();
    if (na!=ma) throw WrongDim("SolveLU(const Matrix, Matrix)", "Matrix must be squared.");
    if (ma!=B.GetM()) throw WrongDim("SolveLu(const Matrix, Matrix)", "Matrices dimensions are incompatible.");
#endif

    // Forward substitution.
    for (i=0; i<ma; i++)
      for (q=0; q<nb; q++)
	{
	  temp=0;
	  for (k=0; k<i; k++)
	    temp+=A(i,k)*B(k,q);
	  B(i,q)=(B(i,q)-temp)/A(i,i);
	}
    // Back substitution.
    for (i=ma-2; i>-1; i--)
      for (q=0; q<nb; q++)
	{
	  temp=0;
	  for (k=i+1; k<ma; k++)
	    temp+=A(i,k)*B(k,q);
	  B(i,q)-=temp;
	}

  }


  // SolveLU //
  /////////////

      
} // namespace Seldon

#define FILE_FUNCTIONS_MATRIX
#endif
