// Copyright (C) 2001-2003 Vivien Mallet
//
// This file is part of Seldon library.
// Seldon library provides matrices and vectors structures for
// linear algebra.
// 
// Seldon is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Seldon is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License (file "license") for more details.
//
// For more information, please see the Seldon home page:
//     http://spacetown.free.fr/lib/seldon/

#ifndef FILE_FUNCTIONS_VECTOR_CXX

/*
  Functions defined in this file:

  alpha.U -> U
  Mlt(alpha, U)

  alpha.U + beta.V -> V
  Add(alpha, U, beta, V)

  alpha.U + beta.V -> W
  Add(alpha, U, beta, V, W)
*/

namespace Seldon
{


  /////////
  // Mlt //


  template <class T0,
	    class T1, class Storage1, class Allocator1>
  void Mlt(const T0 alpha,
	   Vector<T1, Storage1, Allocator1>& A)  throw()
  {
    T1 alpha_=alpha;

    typename Vector<T1, Storage1, Allocator1>::pointer data=A.GetData();

    for (int i=0; i<A.GetDataSize(); i++)
      data[i]=alpha_*data[i];
  }  


  // Mlt //
  /////////



  /////////
  // Add //


  template <class T0,
	    class T1, class Storage1, class Allocator1,
	    class T2, class Storage2, class Allocator2>
  void Add(const T0 alpha,
	   const Vector<T1, Storage1, Allocator1>& A,
	   Vector<T2, Storage2, Allocator2>& C);


  // Add //
  /////////



  /////////
  // Add //


  template <class T0,
	    class T1, class Storage1, class Allocator1,
	    class T2,
	    class T3, class Storage3, class Allocator3>
  void Add(const T0 alpha,
	   const Vector<T1, Storage1, Allocator1>& A,
	   const T2 beta,
	   Vector<T3, Storage3, Allocator3>& C)  throw(WrongDim, NoMemory)
  {
    if ((alpha!=T0(0)) && (beta!=T2(0)))
      {
	T1 alpha_=alpha;
	T3 beta_=beta;

	int ma=A.GetM();
	
#ifdef SELDON_CHECK_BOUNDARIES
	if (ma!=C.GetM()) throw WrongDim("Add(Scalar, const Vector, Scalar, Vector)", "Unable to add vectors.");
#endif

	for (int i=0; i<ma; i++)
	  C(i)=beta_*C(i)+alpha_*A(i);
      }
    else if ((alpha==T0(0)) && (beta==T2(0)))
      {
	C.Zero();
      }
    else if (beta!=T2(0))
      {
	T3 beta_=beta;
	
	typename Vector<T3, Storage3, Allocator3>::pointer data=C.GetData();

	for(int i=0; i<C.GetDataSize(); i++)
	  data[i]=beta_*data[i];
      }
    else
      {
	int ma=A.GetM();
	
	C.Reallocate(ma);

	T1 alpha_=alpha;
	T3 beta_=beta;
	
	for (int i=0; i<ma; i++)
	  C(i)=alpha_*A(i);
      }
  }


  // Add //
  /////////



  /////////
  // Add //


  template <class T0,
	    class T1, class Storage1, class Allocator1,
	    class T2,
	    class T3, class Storage3, class Allocator3,
	    class T4, class Storage4, class Allocator4>
  void Add(const T0 alpha,
	   const Vector<T1, Storage1, Allocator1>& A,
	   const T2 beta,
	   const Vector<T3, Storage3, Allocator3>& B,
	   Vector<T4, Allocator4>& C)  throw(WrongDim, NoMemory)
  {
    if ((alpha!=T0(0)) && (beta!=T2(0)))
      {
	T1 alpha_=alpha;
	T3 beta_=beta;

	int ma=A.GetM();
	
#ifdef SELDON_CHECK_BOUNDARIES
	if (ma!=B.GetM()) throw WrongDim("Add(Scalar, const Vector, Scalar, const Vector, Vector)", "Unable to add vectors.");
#endif

	C.Reallocate(ma);

	for (int i=0; i<ma; i++)
	  C(i)=alpha_*A(i)+beta_*B(i);
      }
    else if ((alpha==T0(0)) && (beta==T2(0)))
      {
	int ma=A.GetM();

	C.Reallocate(ma);
	C.Zero();
      }
    else if (beta!=T3(0))
      {
	int mb=B.GetM();

	T3 beta_=beta;
	
	C.Reallocate(mb);

	for (int i=0; i<mb; i++)
	  C(i)=beta_*B(i);
      }
    else
      {
	int ma=A.GetM();
	
	T1 alpha_=alpha;

	C.Reallocate(ma);

	for (int i=0; i<ma; i++)
	  C(i)=alpha_*A(i);
      }
  }


  // Add //
  /////////

      
} // namespace Seldon

#define FILE_FUNCTIONS_VECTOR_CXX
#endif
