// Copyright (C) 2001-2003 Vivien Mallet
//
// This file is part of Seldon library.
// Seldon library provides matrices and vectors structures for
// linear algebra.
// 
// Seldon is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Seldon is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License (file "license") for more details.
//
// For more information, please see the Seldon home page:
//     http://spacetown.free.fr/lib/seldon/

#ifndef FILE_MATRIX_POINTERS_CXX

#include "Matrix_Pointers.hxx"

namespace Seldon
{


  /****************
   * CONSTRUCTORS *
   ****************/
  

  //! Default constructor.
  /*!
    On exit, the matrix is an empty 0x0 matrix.
  */
  template <class T, class Prop, class Storage, class Allocator>
  inline Matrix_Pointers<T, Prop, Storage, Allocator>::Matrix_Pointers(): Matrix_Base<T, Allocator>()
  {
    me_=NULL;
  }


  //! Main constructor.
  /*! Builds a i x j full matrix.
    \param i number of rows.
    \param j number of columns.
  */
  template <class T, class Prop, class Storage, class Allocator>
  inline Matrix_Pointers<T, Prop, Storage, Allocator>::Matrix_Pointers(int i, int j): Matrix_Base<T, Allocator>(i, j)
  {

#ifdef SELDON_CHECK_MEMORY
    try
      {
#endif

	me_=reinterpret_cast<pointer*>(calloc(Storage::GetFirst(i, j), sizeof(pointer)));

#ifdef SELDON_CHECK_MEMORY
      }
    catch (...)
      {
	this->m_=0;
	this->n_=0;
	me_=NULL;
	this->data_=NULL;
      }
    if (me_==NULL)
      {
	this->m_=0;
	this->n_=0;
	this->data_=NULL;
      }
    if ((me_==NULL) && (i!=0) && (j!=0))
      throw NoMemory("Matrix_Pointers::Matrix_Pointers(int, int)",
		     "Unable to allocate memory for a matrix of size "
		     + to_str(static_cast<long int>(i)
			      * static_cast<long int>(j)
			      * static_cast<long int>(sizeof(T)))
		     + " bytes (" + to_str(i) + " x " + to_str(j)
		     + " elements).");
#endif

#ifdef SELDON_CHECK_MEMORY
    try
      {
#endif

	this->data_=allocator_.allocate(i*j, this);

#ifdef SELDON_CHECK_MEMORY
      }
    catch (...)
      {
	this->m_=0;
	this->n_=0;
	free(me_);
	me_=NULL;
	this->data_=NULL;
      }
    if (data_==NULL)
      {
	this->m_=0;
	this->n_=0;
	free(me_);
	me_=NULL;
      }
    if ((data_==NULL) && (i!=0) && (j!=0))
      throw NoMemory("Matrix_Pointers::Matrix_Pointers(int, int)",
		     "Unable to allocate memory for a matrix of size "
		     + to_str(static_cast<long int>(i)
			      * static_cast<long int>(j)
			      * static_cast<long int>(sizeof(T)))
		     + " bytes (" + to_str(i) + " x " + to_str(j)
		     + " elements).");
#endif

    pointer ptr=this->data_;
    int lgth=Storage::GetSecond(i, j);
    for (int k=0; k<Storage::GetFirst(i, j); k++, ptr+=lgth)
      me_[k]=ptr;
  
  }

  
  /**************
   * DESTRUCTOR *
   **************/


  //! Destructor.
  template <class T, class Prop, class Storage, class Allocator>
  inline Matrix_Pointers<T, Prop, Storage, Allocator>::~Matrix_Pointers()
  {
    
#ifdef SELDON_CHECK_MEMORY
    try
      {
#endif
	
	if (this->data_!=NULL)
	  {
	    allocator_.deallocate(this->data_, this->m_*this->n_);
	    this->data_=NULL;
	  }
	
#ifdef SELDON_CHECK_MEMORY
      }
    catch (...)
      {
	this->data_=NULL;
      }
#endif

#ifdef SELDON_CHECK_MEMORY
    try
      {
#endif

	if (me_!=NULL)
	  {
	    free(me_);
	    me_=NULL;
	  }

#ifdef SELDON_CHECK_MEMORY
      }
    catch (...)
      {
	this->m_=0;
	this->n_=0;
	me_=NULL;
      }
#endif

  }


  //! Clears the matrix.
  /*!
    Frees memory associated with the matrix (i.e. the data array).
    \warning On exit, the matrix is an empty 0x0 matrix.
  */
  template <class T, class Prop, class Storage, class Allocator>
  inline void Matrix_Pointers<T, Prop, Storage, Allocator>::Clear()
  {
    this->~Matrix_Pointers();
    this->m_ = 0;
    this->n_ = 0;
  }


  /*******************
   * BASIC FUNCTIONS *
   *******************/


  //! Returns the number of elements stored in memory.
  /*!
    Returns the number of elements stored in memory, i.e.
    the number of rows multiplied by the number of columns
    because the matrix is full.
    \return The number of elements stored in memory.
  */
  template <class T, class Prop, class Storage, class Allocator>
  int Matrix_Pointers<T, Prop, Storage, Allocator>::GetDataSize() const
  {
    return this->m_*this->n_;
  }


  /*********************
   * MEMORY MANAGEMENT *
   *********************/


  //! Reallocates memory to resize the matrix.
  /*!
    On exit, the matrix is a i x j matrix.
    \param i number of rows.
    \param j number of columns.
    \warning Data may be lost.
  */
  template <class T, class Prop, class Storage, class Allocator>
  inline void Matrix_Pointers<T, Prop, Storage, Allocator>::Reallocate(int i, int j)
  {
    if ((i!=this->m_) || (j!=this->n_))
      {
	this->m_=i;
	this->n_=j;

#ifdef SELDON_CHECK_MEMORY
	try
	  {
#endif

	    me_=reinterpret_cast<pointer*>(realloc(me_, Storage::GetFirst(i, j)*sizeof(pointer)));

#ifdef SELDON_CHECK_MEMORY
	  }
	catch (...)
	  {
	    this->m_=0;
	    this->n_=0;
	    me_=NULL;
	    this->data_=NULL;
	  }
	if (me_==NULL)
	  {
	    this->m_=0;
	    this->n_=0;
	    this->data_=NULL;
	  }
	if ((me_==NULL) && (i!=0) && (j!=0))
	  throw NoMemory("Matrix_Pointers::Reallocate(int, int)",
			 "Unable to reallocate memory for a matrix of size "
			 + to_str(static_cast<long int>(i)
				  * static_cast<long int>(j)
				  * static_cast<long int>(sizeof(T)))
			 + " bytes (" + to_str(i) + " x " + to_str(j)
			 + " elements).");
#endif

#ifdef SELDON_CHECK_MEMORY
	try
	  {
#endif

	    this->data_=reinterpret_cast<pointer>(allocator_.reallocate(this->data_, i*j, this));

#ifdef SELDON_CHECK_MEMORY
	  }
	catch (...)
	  {
	    this->m_=0;
	    this->n_=0;
	    free(me_);
	    me_=NULL;
	    this->data_=NULL;
	  }
	if (this->data_==NULL)
	  {
	    this->m_=0;
	    this->n_=0;
	    free(me_);
	    me_=NULL;
	  }
	if ((data_==NULL) && (i!=0) && (j!=0))
	  throw NoMemory("Matrix_Pointers::Reallocate(int, int)",
			 "Unable to reallocate memory for a matrix of size "
			 + to_str(static_cast<long int>(i)
				  * static_cast<long int>(j)
				  * static_cast<long int>(sizeof(T)))
			 + " bytes (" + to_str(i) + " x " + to_str(j)
			 + " elements).");
#endif

	pointer ptr=this->data_;
	int lgth=Storage::GetSecond(i, j);
	for (int k=0; k<Storage::GetFirst(i, j); k++, ptr+=lgth)
	  me_[k]=ptr;
      }

  }


  //! Changes matrix size and sets its data array.
  /*!
    The matrix is first cleared (memory is freed). The matrix is then set to a i x j matrix,
    and the data array of the matrix is set to 'data'. 'data' elements are not duplicated,
    the new data array of the matrix is the 'data' array, which means that 'data' has to
    be used carefully outside the object. In particular, 'data' will be automatically freed
    when the matrix is destructed.
    \param i number of rows.
    \param j number of columns.
    \param data new array storing elements.
    \warning This method should only be used by advanced users.
  */
  template <class T, class Prop, class Storage, class Allocator>
  inline void Matrix_Pointers<T, Prop, Storage, Allocator>::SetData(int i, int j, typename Matrix_Pointers<T, Prop, Storage, Allocator>::pointer data)
  {

    this->Clear();

    this->m_ = i;
    this->n_ = j;

#ifdef SELDON_CHECK_MEMORY
    try
      {
#endif

	me_=reinterpret_cast<pointer*>(calloc(Storage::GetFirst(i, j), sizeof(pointer)));

#ifdef SELDON_CHECK_MEMORY
      }
    catch (...)
      {
	this->m_=0;
	this->n_=0;
	me_=NULL;
	this->data_=NULL;
	return;
      }
    if (me_==NULL)
      {
	this->m_=0;
	this->n_=0;
	this->data_=NULL;
	return;
      }
#endif

    this->data_=data;

    pointer ptr=this->data_;
    int lgth=Storage::GetSecond(i, j);
    for (int k=0; k<Storage::GetFirst(i, j); k++, ptr+=lgth)
      me_[k]=ptr;

  }


  /**********************************
   * ELEMENT ACCESS AND AFFECTATION *
   **********************************/


  //! Access operator.
  /*!
    Returns the value of element (i, j).
    \param i row index.
    \param j column index.
    \return Element (i, j) of the matrix.
  */
  template <class T, class Prop, class Storage, class Allocator>
  inline typename Matrix_Pointers<T, Prop, Storage, Allocator>::reference Matrix_Pointers<T, Prop, Storage, Allocator>::operator() (int i, int j)
  {

#ifdef SELDON_CHECK_BOUNDARIES
    if ((i<0)||(i>=this->m_)) throw WrongRow("Matrix_Pointers::operator()", "Index should be in [0, "
					     + to_str(this->m_-1) + "], but is equal to "
					     + to_str(i) + ".");
    if ((j<0)||(j>=this->n_)) throw WrongCol("Matrix_Pointers::operator()", "Index should be in [0, "
					     + to_str(this->n_-1) + "], but is equal to "
					     + to_str(j) + ".");
#endif

    return me_[Storage::GetFirst(i, j)][Storage::GetSecond(i, j)];
  }

 
  //! Access operator.
  /*!
    Returns the value of element (i, j).
    \param i row index.
    \param j column index.
    \return Element (i, j) of the matrix.
  */
  template <class T, class Prop, class Storage, class Allocator>
  inline typename Matrix_Pointers<T, Prop, Storage, Allocator>::const_reference Matrix_Pointers<T, Prop, Storage, Allocator>::operator() (int i, int j) const
  {
    
#ifdef SELDON_CHECK_BOUNDARIES
    if ((i<0)||(i>=this->m_)) throw WrongRow("Matrix_Pointers::operator()", "Index should be in [0, "
					     + to_str(this->m_-1) + "], but is equal to "
					     + to_str(i) + ".");
    if ((j<0)||(j>=this->n_)) throw WrongCol("Matrix_Pointers::operator()", "Index should be in [0, "
					     + to_str(this->n_-1) + "], but is equal to "
					     + to_str(j) + ".");
#endif
    
    return me_[Storage::GetFirst(i, j)][Storage::GetSecond(i, j)];
  }


  //! Copies a given matrix.
  /*!
    Sets the current matrix to 'A'.
    \param A matrix to be copied.
    \note Matrix elements are shared.
  */
  template <class T, class Prop, class Storage, class Allocator>
  inline void Matrix_Pointers<T, Prop, Storage, Allocator>::Copy(const Matrix_Pointers<T, Prop, Storage, Allocator>& A)
  {
    this->Reallocate(A.GetM(), A.GetN());

    allocator_.memorycpy(this->data_, A.GetData(), this->GetDataSize());
  }


  /************************
   * CONVENIENT FUNCTIONS *
   ************************/


  //! Sets all elements to zero.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::Zero()
  {
    allocator_.memoryset(this->data_, char(0), this->GetDataSize()*sizeof(value_type));
  }


  //! Sets the current matrix to the identity.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::SetIdentity()
  {
    allocator_.memoryset(this->data_, char(0), this->GetDataSize()*sizeof(value_type));
    T one(1);
    for (int i=0; i<min(this->m_, this->n_); i++)
      (*this)(i,i)=one;
  }


  //! Fills a matrix.
  /*!
    On exit, the matrix is filled with 1, 2, 3, 4, ... The order of
    those numbers depends on the storage.
  */
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::Fill()
  {
    for (int i=0; i<this->GetDataSize(); i++)
      this->data_[i]=i+1;
  }


  //! Fills a matrix with a given value.
  /*!
    On exit, the matrix is filled with 'x'.
    \param x the value to fill the matrix with.
  */
  template <class T, class Prop, class Storage, class Allocator>
  template <class T0>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::Fill(T0 x)
  {
    for (int i=0; i<this->GetDataSize(); i++)
      this->data_[i]=x;
  }


  //! Fills a matrix randomly.
  /*!
    On exit, the matrix is filled with random values.
  */
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::FillRand()
  {
    srand(time(NULL));
    for (int i=0; i<this->GetDataSize(); i++)
      this->data_[i]=rand();
  }


  //! Displays the matrix on the standard output.
  /*!
    Displays elements on the standard output, in text format.
    Each row is displayed on a single line and elements of
    a row are delimited by tabulations.
  */
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::Print() const
  {
    for (int i=0; i<this->m_; i++)
      {
	for (int j=0; j<this->n_; j++)
	  cout << (*this)(i,j) << "\t";
	cout << endl;
      }
  }


  //! Displays a sub-matrix on the standard output.
  /*!
    The sub-matrix is defined by its upper-left corner (a, b)
    and its bottom-right corner (m, n). So, elements with indices
    in [a, m] x [b, n] are displayed on the standard output,
    in text format. Each row is displayed on a single line and
    elements of a row are delimited by tabulations.
    \param a row index of the upper-left corner.
    \param b column index of the upper-left corner.
    \param m row index of the bottom-right corner.
    \param n column index of the bottom-right corner.
  */
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::Print(int a, int b, int m, int n) const
  {
    for (int i=a; i<min(this->m_, a+m); i++)
      {
	for (int j=b; j<min(this->n_, b+n); j++)
	  cout << (*this)(i,j) << "\t";
	cout << endl;
      }
  }


  //! Displays a square sub-matrix on the standard output.
  /*!
    The sub-matrix is defined by its bottom-right corner (l, l).
    So, elements with indices in [0, 0] x [l, l] are displayed
    on the standard output, in text format. Each row is displayed
    on a single line and elements of a row are delimited
    by tabulations.
    \param l dimension of the square matrix to be displayed.
  */
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::Print(int l) const
  {
    Print(0, 0, l, l);
  }


  /*********
   * NORMS *
   *********/


  //! Returns the maximum (in absolute value) of the matrix.
  /*!
    \return The maximum (in absolute value) of the matrix.
    \note The name of this method is of course not relevant
    since the infinity norm of the matrix is something else.
    The name of this method will be GetMaxAbs in a next version.
  */
  template <class T, class Prop, class Storage, class Allocator>
  typename Matrix_Pointers<T, Prop, Storage, Allocator>::value_type Matrix_Pointers<T, Prop, Storage, Allocator>::GetNormInf() const
  {
    value_type res=value_type(0);
    int i, j;
    for (i=0; i<this->GetM(); i++)
      for (j=0; j<this->GetN(); j++)
	{
	  res=max(res, (*this)(i,j));
	  res=max(res, -(*this)(i,j));
	}

    return res;
  }


  /**************************
   * INPUT/OUTPUT FUNCTIONS *
   **************************/


  //! Writes the matrix in a file.
  /*!
    Stores the matrix in a file in binary format.
    The number of rows (integer) and the number of columns (integer)
    are written, and matrix elements are then written in the same order
    as in memory (e.g. row-major storage).
    \param FileName output file name.
  */
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::Write(string FileName) const
  {

    ofstream FileStream;
    FileStream.open(FileName.c_str());

#ifdef SELDON_CHECK_IO
    // Checks if the file was opened.
    if (!FileStream.is_open())
      throw IOError("Matrix_Pointers::Write(string FileName)",
		    "Unable to open file \"" + FileName + "\".");
#endif

    this->Write(FileStream);

    FileStream.close();
    
  }


  //! Writes the matrix to an output stream.
  /*!
    Writes the matrix to an output stream in binary format.
    The number of rows (integer) and the number of columns (integer)
    are written, and matrix elements are then written in the same order
    as in memory (e.g. row-major storage).
    \param FileStream output stream.
  */
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::Write(ofstream& FileStream) const
  {

#ifdef SELDON_CHECK_IO
    // Checks if the stream is ready.
    if (!FileStream.good())
      throw IOError("Matrix_Pointers::Write(ofstream& FileStream)",
		    "Stream is not ready.");
#endif

    FileStream.write(reinterpret_cast<char*>(const_cast<int*>(&this->m_)),
		     sizeof(int));
    FileStream.write(reinterpret_cast<char*>(const_cast<int*>(&this->n_)),
		     sizeof(int));

    FileStream.write(reinterpret_cast<char*>(this->data_),
		     this->m_ * this->n_ * sizeof(value_type));

#ifdef SELDON_CHECK_IO
    // Checks if data was written.
    if (!FileStream.good())
      throw IOError("Matrix_Pointers::Write(ofstream& FileStream)",
                    string("Output operation failed. The output file may have been removed")
		    + "or there is no space left on device.");
#endif

  }


  //! Writes the matrix in a file.
  /*!
    Stores the matrix in a file in text format.
    Only matrix elements are written (not dimensions).
    Each row is written on a single line and elements of
    a row are delimited by tabulations.
    \param FileName output file name.
  */
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::WriteText(string FileName) const
  {

    ofstream FileStream;
    FileStream.open(FileName.c_str());

#ifdef SELDON_CHECK_IO
    // Checks if the file was opened.
    if (!FileStream.is_open())
      throw IOError("Matrix_Pointers::WriteText(string FileName)",
		    "Unable to open file \"" + FileName + "\".");
#endif

    this->WriteText(FileStream);

    FileStream.close();

  }


  //! Writes the matrix to an output stream.
  /*!
    Writes the matrix to an output stream in text format.
    Only matrix elements are written (not dimensions).
    Each row is written on a single line and elements of
    a row are delimited by tabulations.
    \param FileStream output stream.
  */
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::WriteText(ofstream& FileStream) const
  {

#ifdef SELDON_CHECK_IO
    // Checks if the file is ready.
    if (!FileStream.good())
      throw IOError("Matrix_Pointers::WriteText(ofstream& FileStream)",
                    "Stream is not ready.");
#endif

    int i, j;
    for (i=0; i<this->GetM(); i++)
      {
	for (j=0; j<this->GetN(); j++)
	  FileStream << (*this)(i, j) << '\t';
	FileStream << endl;
      }

#ifdef SELDON_CHECK_IO
    // Checks if data was written.
    if (!FileStream.good())
      throw IOError("Matrix_Pointers::WriteText(ofstream& FileStream)",
                    string("Output operation failed. The output file may have been removed")
		    + "or there is no space left on device.");
#endif

  }


  //! Reads the matrix from a file.
  /*!
    Reads a matrix stored in binary format in a file.
    The number of rows (integer) and the number of columns (integer)
    are read, and matrix elements are then read in the same order
    as it should be in memory (e.g. row-major storage).
    \param FileName input file name.
  */
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::Read(string FileName)
  {

    ifstream FileStream;
    FileStream.open(FileName.c_str());

#ifdef SELDON_CHECK_IO
    // Checks if the file was opened.
    if (!FileStream.is_open())
      throw IOError("Matrix_Pointers::Read(string FileName)",
		    "Unable to open file \"" + FileName + "\".");
#endif

    this->Read(FileStream);

    FileStream.close();

  }
 

  //! Reads the matrix from an input stream.
  /*!
    Reads a matrix in binary format from an input stream.
    The number of rows (integer) and the number of columns (integer)
    are read, and matrix elements are then read in the same order
    as it should be in memory (e.g. row-major storage).
    \param FileStream input stream.
  */
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_Pointers<T, Prop, Storage, Allocator>::Read(ifstream& FileStream)
  {

#ifdef SELDON_CHECK_IO
    // Checks if the stream is ready.
    if (!FileStream.good())
      throw IOError("Matrix_Pointers::Read(ifstream& FileStream)",
                    "Stream is not ready.");
#endif

    int new_m, new_n;
    FileStream.read(reinterpret_cast<char*>(&new_m), sizeof(int));
    FileStream.read(reinterpret_cast<char*>(&new_n), sizeof(int));
    this->Reallocate(new_m, new_n);

    FileStream.read(reinterpret_cast<char*>(this->data_),
		    new_m * new_n * sizeof(value_type));

#ifdef SELDON_CHECK_IO
    // Checks if data was read.
    if (!FileStream.good())
      throw IOError("Matrix_Pointers::Read(ifstream& FileStream)",
                    string("Output operation failed.")
		    + " The intput file may have been removed or may not contain enough data.");
#endif    

  }



  //////////////////////
  // MATRIX<COLMAJOR> //
  //////////////////////


  /****************
   * CONSTRUCTORS *
   ****************/


  //! Default constructor.
  /*!
    Builds a empty matrix.
  */
  template <class T, class Prop, class Allocator>
  Matrix<T, Prop, ColMajor, Allocator>::Matrix()  throw():
    Matrix_Pointers<T, Prop, ColMajor, Allocator>()
  {
  }


  //! Main constructor.
  /*! Builds a i by j full column-major matrix.
    \param i number of rows.
    \param j number of columns.
  */
  template <class T, class Prop, class Allocator>
  Matrix<T, Prop, ColMajor, Allocator>::Matrix(int i, int j):
    Matrix_Pointers<T, Prop, ColMajor, Allocator>(i, j)
  {
  }



  //////////////////////
  // MATRIX<ROWMAJOR> //
  //////////////////////


  /****************
   * CONSTRUCTORS *
   ****************/


  //! Default constructor.
  /*!
    Builds a empty matrix.
  */
  template <class T, class Prop, class Allocator>
  Matrix<T, Prop, RowMajor, Allocator>::Matrix()  throw():
    Matrix_Pointers<T, Prop, RowMajor, Allocator>()
  {
  }


  //! Main constructor.
  /*! Builds a i by j full row-major matrix.
    \param i number of rows.
    \param j number of columns.
  */
  template <class T, class Prop, class Allocator>
  Matrix<T, Prop, RowMajor, Allocator>::Matrix(int i, int j):
    Matrix_Pointers<T, Prop, RowMajor, Allocator>(i, j)
  {
  }


} // namespace Seldon

#define FILE_MATRIX_POINTERS_CXX
#endif
