// Copyright (C) 2001-2003 Vivien Mallet
//
// This file is part of Seldon library.
// Seldon library provides matrices and vectors structures for
// linear algebra.
// 
// Seldon is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Seldon is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License (file "license") for more details.
//
// For more information, please see the Seldon home page:
//     http://spacetown.free.fr/lib/seldon/

#ifndef FILE_MATRIX_SYMPACKED_CXX

#include "Matrix_SymPacked.hxx"

namespace Seldon
{


  /****************
   * CONSTRUCTORS *
   ****************/
  

  // Default constructor.
  template <class T, class Prop, class Storage, class Allocator>
  inline Matrix_SymPacked<T, Prop, Storage, Allocator>::Matrix_SymPacked(): Matrix_Base<T, Allocator>()
  {
  }


  // Constructor.
  template <class T, class Prop, class Storage, class Allocator>
  inline Matrix_SymPacked<T, Prop, Storage, Allocator>::Matrix_SymPacked(int i, int j): Matrix_Base<T, Allocator>(i, i)
  {

#ifdef SELDON_CHECK_MEMORY
    try
      {
#endif

	this->data_=allocator_.allocate((i*(i+1))/2, this);

#ifdef SELDON_CHECK_MEMORY
      }
    catch (...)
      {
	this->m_=0;
	this->n_=0;
	this->data_=NULL;
	return;
      }
    if (data_==NULL)
      {
	this->m_=0;
	this->n_=0;
	return;
      }
#endif

  }

  
  /**************
   * DESTRUCTOR *
   **************/


  // Destructor.
  template <class T, class Prop, class Storage, class Allocator>
  inline Matrix_SymPacked<T, Prop, Storage, Allocator>::~Matrix_SymPacked()
  {
    
#ifdef SELDON_CHECK_MEMORY
    try
      {
#endif
	
	if (this->data_!=NULL)
	  {
	    allocator_.deallocate(this->data_, (this->m_*(this->m_+1))/2);
	    this->data_=NULL;
	  }
	
#ifdef SELDON_CHECK_MEMORY
      }
    catch (...)
      {
	this->m_=0;
	this->n_=0;
	this->data_=NULL;
      }
#endif

  }


  // "Destructor".
  template <class T, class Prop, class Storage, class Allocator>
  inline void Matrix_SymPacked<T, Prop, Storage, Allocator>::Clear()
  {
    this->~Matrix_SymPacked();
  }


  /*******************
   * BASIC FUNCTIONS *
   *******************/


  // Returns the number of stored elements.
  template <class T, class Prop, class Storage, class Allocator>
  int Matrix_SymPacked<T, Prop, Storage, Allocator>::GetDataSize() const
  {
    return (this->m_*(this->m_+1))/2;
  }


  /*********************
   * MEMORY MANAGEMENT *
   *********************/


  // Reallocation (data lost).
  template <class T, class Prop, class Storage, class Allocator>
  inline void Matrix_SymPacked<T, Prop, Storage, Allocator>::Reallocate(int i, int j)
  {
    if (i!=this->m_)
      {
	this->m_=i;
	this->n_=i;

#ifdef SELDON_CHECK_MEMORY
	try
	  {
#endif

	    this->data_=reinterpret_cast<pointer>(allocator_.reallocate(this->data_, (i*(i+1))/2, this));

#ifdef SELDON_CHECK_MEMORY
	  }
	catch (...)
	  {
	    this->m_=0;
	    this->n_=0;
	    this->data_=NULL;
	    throw NoMemory("Matrix_SymPacked::Reallocate(int, int)", "Unable to reallocate memory for data_.");
	  }
	if (this->data_==NULL)
	  {
	    this->m_=0;
	    this->n_=0;
	    throw NoMemory("Matrix_SymPacked::Reallocate(int, int)", "Unable to reallocate memory for data_.");
	  }
#endif

      }

  }


  // Sets data.
  template <class T, class Prop, class Storage, class Allocator>
  inline void Matrix_SymPacked<T, Prop, Storage, Allocator>::SetData(int i, int j, typename Matrix_SymPacked<T, Prop, Storage, Allocator>::pointer data)
  {

    this->Clear();

    this->m_ = i;
    this->n_ = i;

    this->data_ = data;

  }


  /**********************************
   * ELEMENT ACCESS AND AFFECTATION *
   **********************************/


  // Allows to access to elements.
  template <class T, class Prop, class Storage, class Allocator>
  inline typename Matrix_SymPacked<T, Prop, Storage, Allocator>::reference Matrix_SymPacked<T, Prop, Storage, Allocator>::operator() (int i, int j)
  {

#ifdef SELDON_CHECK_BOUNDARIES
    if ((i<0)||(i>=this->m_)) throw WrongRow("Matrix_SymPacked::operator()", "Index should be in [0, "
					     + to_str(this->m_-1) + "], but is equal to "
					     + to_str(i) + ".");
    if ((j<0)||(j>=this->n_)) throw WrongCol("Matrix_SymPacked::operator()", "Index should be in [0, "
					     + to_str(this->n_-1) + "], but is equal to "
					     + to_str(j) + ".");
#endif

    return data_[j>i?Storage::GetFirst(i * this->n_ - (i*(i+1)) / 2 + j, (j*(j+1)) / 2 + i):Storage::GetFirst(j * this->m_ - (j*(j+1)) / 2 + i, (i*(i+1)) / 2 + j)];
  }

 
  // Allows to access to elements.
  template <class T, class Prop, class Storage, class Allocator>
  inline typename Matrix_SymPacked<T, Prop, Storage, Allocator>::const_reference Matrix_SymPacked<T, Prop, Storage, Allocator>::operator() (int i, int j) const
  {
    
#ifdef SELDON_CHECK_BOUNDARIES
    if ((i<0)||(i>=this->m_)) throw WrongRow("Matrix_SymPacked::operator()", "Index should be in [0, "
					     + to_str(this->m_-1) + "], but is equal to "
					     + to_str(i) + ".");
    if ((j<0)||(j>=this->n_)) throw WrongCol("Matrix_SymPacked::operator()", "Index should be in [0, "
					     + to_str(this->n_-1) + "], but is equal to "
					     + to_str(j) + ".");
#endif
    
    return data_[j>i?Storage::GetFirst(i * this->n_ - (i*(i+1)) / 2 + j, (j*(j+1)) / 2 + i):Storage::GetFirst(j * this->m_ - (j*(j+1)) / 2 + i, (i*(i+1)) / 2 + j)];
  }


  // Copy a matrix to another.
  template <class T, class Prop, class Storage, class Allocator>
  inline void Matrix_SymPacked<T, Prop, Storage, Allocator>::Copy(const Matrix_SymPacked<T, Prop, Storage, Allocator>& A)
  {
    this->Reallocate(A.GetM(), A.GetN());

    allocator_.memorycpy(this->data_, A.GetData(), this->GetDataSize());
  }


  /************************
   * CONVENIENT FUNCTIONS *
   ************************/


  // Sets all elements to zero.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::Zero()
  {
    allocator_.memoryset(this->data_, char(0), this->GetDataSize()*sizeof(value_type));
  }


  // Sets the current matrix to the identity.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::SetIdentity()
  {
    allocator_.memoryset(this->data_, char(0), this->GetDataSize()*sizeof(value_type));
    T one(1);
    for (int i=0; i<min(this->m_, this->n_); i++)
      (*this)(i,i)=one;
  }


  // Fills a matrix.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::Fill()
  {
    for (int i=0; i<this->GetDataSize(); i++)
      this->data_[i]=i+1;
  }


  // Fills a matrix with a given value.
  template <class T, class Prop, class Storage, class Allocator>
  template <class T0>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::Fill(T0 x)
  {
    for (int i=0; i<this->GetDataSize(); i++)
      this->data_[i]=x;
  }


  // Fills a matrix randomly.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::FillRand()
  {
    srand(time(NULL));
    for (int i=0; i<this->GetDataSize(); i++)
      this->data_[i]=rand();
  }


  // Displays a matrix.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::Print() const
  {
    for (int i=0; i<this->m_; i++)
      {
	for (int j=0; j<this->n_; j++)
	  cout << (*this)(i,j) << "\t";
	cout << endl;
      }
  }


  // Displays a sub-matrix.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::Print(int a, int b, int m, int n) const
  {
    for (int i=a; i<min(this->m_, a+m); i++)
      {
	for (int j=b; j<min(this->n_, b+n); j++)
	  cout << (*this)(i,j) << "\t";
	cout << endl;
      }
  }


  // Displays a matrix.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::Print(int l) const
  {
    Print(0, 0, l, l);
  }


  /*********
   * NORMS *
   *********/

  // A refaire...
  // Returns infinite (?) norm.
  template <class T, class Prop, class Storage, class Allocator>
  typename Matrix_SymPacked<T, Prop, Storage, Allocator>::value_type Matrix_SymPacked<T, Prop, Storage, Allocator>::GetNormInf() const
  {
    value_type res=value_type(0);
    int i, j;
    for (i=0; i<this->GetM(); i++)
      for (j=0; j<this->GetN(); j++)
	{
	  res=max(res, (*this)(i,j));
	  res=max(res, -(*this)(i,j));
	}

    return res;
  }


  /**************************
   * INPUT/OUTPUT FUNCTIONS *
   **************************/


  // Writes in a file.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::Write(string FileName) const
  {

    ofstream FileStream;
    FileStream.open(FileName.c_str());

#ifdef SELDON_CHECK_IO
    // Checks if the file was opened.
    if (!FileStream.is_open())
      throw IOError("Matrix_SymPacked::Write(string FileName)",
		    "Unable to open file \"" + FileName + "\".");
#endif

    this->Write(FileStream);

    FileStream.close();
    
  }


  // Writes in a file.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::Write(ofstream& FileStream) const
  {
    
#ifdef SELDON_CHECK_IO
    // Checks if the file is ready.
    if (!FileStream.good())
      throw IOError("Matrix_SymPacked::Write(ofstream& FileStream)",
                    "Stream is not ready.");
#endif

    FileStream.write(reinterpret_cast<char*>(const_cast<int*>(&this->m_)),
		     sizeof(int));
    FileStream.write(reinterpret_cast<char*>(const_cast<int*>(&this->n_)),
		     sizeof(int));

    FileStream.write(reinterpret_cast<char*>(this->data_),
		     this->GetDataSize() * sizeof(value_type));

#ifdef SELDON_CHECK_IO
    // Checks if data was written.
    if (!FileStream.good())
      throw IOError("Matrix_SymPacked::Write(ofstream& FileStream)",
                    string("Output operation failed. The output file may have been removed")
		    + "or there is no space left on device.");
#endif

  }


  // Writes in a file.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::WriteText(string FileName) const
  {

    ofstream FileStream;
    FileStream.open(FileName.c_str());

#ifdef SELDON_CHECK_IO
    // Checks if the file was opened.
    if (!FileStream.is_open())
      throw IOError("Matrix_SymPacked::WriteText(string FileName)",
		    "Unable to open file \"" + FileName + "\".");
#endif

    this->WriteText(FileStream);

    FileStream.close();

  }


  // Writes in a file.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::WriteText(ofstream& FileStream) const
  {

#ifdef SELDON_CHECK_IO
    // Checks if the stream is ready.
    if (!FileStream.good())
      throw IOError("Matrix_SymPacked::WriteText(ofstream& FileStream)",
                    "Stream is not ready.");
#endif

    int i, j;
    for (i=0; i<this->GetM(); i++)
      {
	for (j=0; j<this->GetN(); j++)
	  FileStream << (*this)(i, j) << '\t';
	FileStream << endl;
      }

#ifdef SELDON_CHECK_IO
    // Checks if data was written.
    if (!FileStream.good())
      throw IOError("Matrix_SymPacked::WriteText(ofstream& FileStream)",
                    string("Output operation failed. The output file may have been removed")
		    + "or there is no space left on device.");
#endif

  }


  // Reads a file.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::Read(string FileName)
  {

    ifstream FileStream;
    FileStream.open(FileName.c_str());

#ifdef SELDON_CHECK_IO
    // Checks if the file was opened.
    if (!FileStream.good())
      throw IOError("Matrix_SymPacked::Read(string FileName)",
		    "Unable to open file \"" + FileName + "\".");
#endif

    this->Read(FileStream);

    FileStream.close();

  }
 

  // Reads a file.
  template <class T, class Prop, class Storage, class Allocator>
  void Matrix_SymPacked<T, Prop, Storage, Allocator>::Read(ifstream& FileStream)
  {

#ifdef SELDON_CHECK_IO
    // Checks if the stream is ready.
    if (!FileStream.good())
      throw IOError("Matrix_SymPacked::Read(ifstream& FileStream)",
                    "Stream is not ready.");
#endif

    int new_m, new_n;
    FileStream.read(reinterpret_cast<char*>(&new_m), sizeof(int));
    FileStream.read(reinterpret_cast<char*>(&new_n), sizeof(int));
    this->Reallocate(new_m, new_n);

    FileStream.read(reinterpret_cast<char*>(this->data_),
		    this->GetDataSize() * sizeof(value_type));

#ifdef SELDON_CHECK_IO
    // Checks if data was read.
    if (!FileStream.good())
      throw IOError("Matrix_SymPacked::Read(ifstream& FileStream)",
                    string("Output operation failed.")
		    + " The intput file may have been removed or may not contain enough data.");
#endif    

  }


} // namespace Seldon

#define FILE_MATRIX_SYMPACKED_CXX
#endif
