// Copyright (C) 2001-2003 Vivien Mallet
//
// This file is part of Seldon library.
// Seldon library provides matrices and vectors structures for
// linear algebra.
// 
// Seldon is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Seldon is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License (file "license") for more details.
//
// For more information, please see the Seldon home page:
//     http://spacetown.free.fr/lib/seldon/

#ifndef FILE_VECTOR_CXX

#include "Vector.hxx"

namespace Seldon
{


  ///////////////////
  // C_VECTOR_BASE //
  ///////////////////


  /****************
   * CONSTRUCTORS *
   ****************/
  

  // Default constructor.
  template <class T, class Allocator>
  inline Vector_Base<T, Allocator>::Vector_Base()
  {
    m_=0;
    data_=NULL;
  }


  // Constructor.
  template <class T, class Allocator>
  inline Vector_Base<T, Allocator>::Vector_Base(int i)
  {
    m_=i;
  }


  // Copy constructor.
  template <class T, class Allocator>
  inline Vector_Base<T, Allocator>::Vector_Base(Vector_Base<T, Allocator>& A)
  {
    m_=A.GetM();
  }


  /**************
   * DESTRUCTOR *
   **************/
  

  // Destructor.
  template <class T, class Allocator>
  inline Vector_Base<T, Allocator>::~Vector_Base()
  {

#ifdef SELDON_CHECK_MEMORY
    try
      {
#endif

	if (data_!=NULL)
	  {
	    vect_allocator_.deallocate(data_, m_);
	    m_ = 0;
	    data_=NULL;
	  }
	
#ifdef SELDON_CHECK_MEMORY
      }
    catch (...)
      {
	m_=0;
	data_=NULL;
      }
#endif

  }


  /*******************
   * BASIC FUNCTIONS *
   *******************/
  

  // Returns the number of stored elements.
  template <class T, class Allocator>
  int Vector_Base<T, Allocator>::GetM() const
  {
    return m_;
  }


  // Returns the number of stored elements.
  template <class T, class Allocator>
  int Vector_Base<T, Allocator>::GetLength() const
  {
    return m_;
  }


  // Returns the number of stored elements.
  template <class T, class Allocator>
  int Vector_Base<T, Allocator>::GetSize() const
  {
    return m_;
  }


  // Returns a pointer to data_ (stored data).
  template <class T, class Allocator>
  typename Vector_Base<T, Allocator>::pointer Vector_Base<T, Allocator>::GetData() const
  {
    return data_;
  }



  ///////////////////////
  // VECTOR<VECT_FULL> //
  ///////////////////////

  
  /****************
   * CONSTRUCTORS *
   ****************/


  //! Default constructor.
  /*!
    On exit, the vector is empty.
  */
  template <class T, class Allocator>
  Vector<T, Vect_Full, Allocator>::Vector()  throw(): Vector_Base<T, Allocator>()
  {
  }


  //! Main constructor.
  /*! Builds a vector of a given size.
    \param i vector size.
  */
  template <class T, class Allocator>
  Vector<T, Vect_Full, Allocator>::Vector(int i):
    Vector_Base<T, Allocator>(i)
  {

#ifdef SELDON_CHECK_MEMORY
    try
      {
#endif
	  
	this->data_=vect_allocator_.allocate(i, this);
	  
#ifdef SELDON_CHECK_MEMORY
      }
    catch (...)
      {
	this->m_=0;
	this->data_=NULL;
      }
    if (data_==NULL)
      this->m_=0;
    if ( (data_==NULL) && (i!=0))
      throw NoMemory("Vector<Vect_Full>::Vector(int)",
		     "Unable to allocate memory for a vector of size "
		     + to_str(i*sizeof(T)) + " bytes ("
		     + to_str(i) + " elements).");
#endif
      
  }



  //! Copy constructor.
  /*! Builds a copy of a vector.
    \param V vector to be copied.
  */
  template <class T, class Allocator>
  Vector<T, Vect_Full, Allocator>::Vector(Vector<T, Vect_Full, Allocator>& V):
    Vector_Base<T, Allocator>(V)
  {
      
#ifdef SELDON_CHECK_MEMORY
    try
      {
#endif
	  
	this->data_=vect_allocator_.allocate(V.GetM(), this);

#ifdef SELDON_CHECK_MEMORY
      }
    catch (...)
      {
	this->m_=0;
	this->data_=NULL;
      }
    if (data_==NULL)
      this->m_=0;
    if ((data_==NULL) && (V.GetM()!=0))
      throw NoMemory("Vector<Vect_Full>::Vector(Vector<Vect_Full>&)",
		     "Unable to allocate memory for a vector of size "
		     + to_str(V.GetM()*sizeof(T)) + " bytes ("
		     + to_str(V.GetM()) + " elements).");
#endif

    vect_allocator_.memorycpy(this->data_, V.GetData(), V.GetM());
	  
  }

  
  /**************
   * DESTRUCTOR *
   **************/


  //! Destructor.
  template <class T, class Allocator>
  Vector<T, Vect_Full, Allocator>::~Vector()
  {
  }


  /*********************
   * MEMORY MANAGEMENT *
   *********************/


  // "Destructor".
  template <class T, class Allocator>
  inline void Vector<T, Vect_Full, Allocator>::Clear()
  {
    this->~Vector();
  }


  // Reallocation.
  template <class T, class Allocator>
  inline void Vector<T, Vect_Full, Allocator>::Reallocate(int i)
  {
    if (i!=this->m_)
      {

	this->m_=i;

#ifdef SELDON_CHECK_MEMORY
	try
	  {
#endif

	    this->data_=reinterpret_cast<pointer>(vect_allocator_.reallocate(this->data_, i, this));

#ifdef SELDON_CHECK_MEMORY
	  }
	catch (...)
	  {
	    this->m_=0;
	    this->data_=NULL;
	    return;
	  }
	if (this->data_==NULL)
	  {
	    this->m_=0;
	    return;
	  }
#endif

      }

  }


  // Reallocation.
  template <class T, class Allocator>
  inline void Vector<T, Vect_Full, Allocator>::SetData(int i, typename Vector<T, Vect_Full, Allocator>::pointer data)
  {

    this->Clear();

    this->m_=i;

    this->data_=data;

  }


  //! Clears the vector without releasing memory.
  /*!
    On exit, the vector is empty and the memory has not been released.
    It is useful for low level manipulations on a Vector instance.
    \warning Memory is not released.
  */
  template <class T, class Allocator>
  void Vector<T, Vect_Full, Allocator>::Nullify()
  {
    this->m_=0;
    this->data_=NULL;
  }


  /**********************************
   * ELEMENT ACCESS AND AFFECTATION *
   **********************************/


  // Allows to access to elements.
  template <class T, class Allocator>
  inline typename Vector<T, Vect_Full, Allocator>::reference Vector<T, Vect_Full, Allocator>::operator() (int i)
  {

#ifdef SELDON_CHECK_BOUNDARIES
    if ((i<0)||(i>=this->m_)) throw WrongIndex("Vector<Vect_Full>::operator()", "Index should be in [0, "
					       + to_str(this->m_-1) + "], but is equal to "
					       + to_str(i) + ".");
#endif

    return this->data_[i];
  }


  // Allows to access to elements.
  template <class T, class Allocator>
  inline typename Vector<T, Vect_Full, Allocator>::const_reference Vector<T, Vect_Full, Allocator>::operator() (int i) const
  {

#ifdef SELDON_CHECK_BOUNDARIES
    if ((i<0)||(i>=this->m_)) throw WrongIndex("Vector<Vect_Full>::operator()", "Index should be in [0, "
					       + to_str(this->m_-1) + "], but is equal to "
					       + to_str(i) + ".");
#endif

    return this->data_[i];
  }


  // Copy a vector to another.
  template <class T, class Allocator>
  inline void Vector<T, Vect_Full, Allocator>::Copy(const Vector<T, Vect_Full, Allocator>& X)
  {
    this->Reallocate(X.GetLength());

    vect_allocator_.memorycpy(this->data_, X.GetData(), this->m_);
  }


  /*******************
   * BASIC FUNCTIONS *
   *******************/

  // Returns the number of stored elements.
  template <class T, class Allocator>
  int Vector<T, Vect_Full, Allocator>::GetDataSize()
  {
    return this->m_;
  }


  /************************
   * CONVENIENT FUNCTIONS *
   ************************/


  // Sets all elements to zero.
  template <class T, class Allocator>
  void Vector<T, Vect_Full, Allocator>::Zero()
  {
    vect_allocator_.memoryset(this->data_, char(0), this->GetDataSize()*sizeof(value_type));
  }


  // Fills the current vector.
  template <class T, class Allocator>
  void Vector<T, Vect_Full, Allocator>::Fill()
  {
    for (int i=0; i<this->m_; i++)
      this->data_[i]=i+1;
  }


  // Fills the current vector with a given value.
  template <class T, class Allocator>
  template <class T0>
  void Vector<T, Vect_Full, Allocator>::Fill(T0 x)
  {
    for (int i=0; i<this->m_; i++)
      this->data_[i]=x;
  }


  // Fills a vector randomly.
  template <class T, class Allocator>
  void Vector<T, Vect_Full, Allocator>::FillRand()
  {
    srand(time(NULL));
    for (int i=0; i<this->m_; i++)
      this->data_[i]=rand();
  }


  // Displays the current vector.
  template <class T, class Allocator>
  void Vector<T, Vect_Full, Allocator>::Print() const
  {
    for (int i=0; i<this->GetLength(); i++)
      cout << (*this)(i) << "\t";
    cout << endl;
  }


  /*********
   * NORMS *
   *********/

  // A refaire...
  // Returns infinite (?) norm.
  template <class T, class Allocator>
  typename Vector<T, Vect_Full, Allocator>::value_type Vector<T, Vect_Full, Allocator>::GetNormInf() const
  {
    value_type res=value_type(0);
    for (int i=0; i<this->GetLength(); i++)
      {
	res=max(res, this->data_[i]);
	res=max(res, -(this->data_[i]));
      }

    return res;
  }


  // Return the index of the highest absolute value.
  template <class T, class Allocator>
  int Vector<T, Vect_Full, Allocator>::GetNormInfIndex() const
  {

#ifdef SELDON_CHECK_DIMENSIONS
    if (this->GetLength()==0) throw WrongDim("Vector<Vect_Full>::GetNormInfIndex()", "Vector is null.");
#endif

    value_type res=value_type(0), temp;
    int j=0;
    for (int i=0; i<this->GetLength(); i++)
      {
	temp=res;
	res=max(res, this->data_[i]);
	res=max(res, -(this->data_[i]));
	if (temp!=res) j=i;
      }

    return j;
  }


  /**************************
   * OUTPUT/INPUT FUNCTIONS *
   **************************/


  // Writes in a file.
  template <class T, class Allocator>
  void Vector<T, Vect_Full, Allocator>::Write(string FileName) const
  {

    ofstream FileStream;
    FileStream.open(FileName.c_str());

#ifdef SELDON_CHECK_IO
    // Checks if the file was opened.
    if (!FileStream.is_open())
      throw IOError("Vector<Vect_Full>::Write(string FileName)",
		    "Unable to open file \"" + FileName + "\".");
#endif

    this->Write(FileStream);
    
    FileStream.close();
    
  }


  // Writes in a file.
  template <class T, class Allocator>
  void Vector<T, Vect_Full, Allocator>::Write(ofstream& FileStream) const
  {

#ifdef SELDON_CHECK_IO
    // Checks if the stream is ready.
    if (!FileStream.good())
      throw IOError("Vector<Vect_Full>::Write(ofstream& FileStream)",
                    "Stream is not ready.");
#endif

    FileStream.write(reinterpret_cast<char*>(const_cast<int*>(&this->m_)),
		     sizeof(int));

    FileStream.write(reinterpret_cast<char*>(this->data_),
		     this->m_ * sizeof(value_type));

#ifdef SELDON_CHECK_IO
    // Checks if data was written.
    if (!FileStream.good())
      throw IOError("Vector<Vect_Full>::Write(ofstream& FileStream)",
                    string("Output operation failed. The output file may have been removed")
		    + "or there is no space left on device.");
#endif

  }


  // Writes in a file.
  template <class T, class Allocator>
  void Vector<T, Vect_Full, Allocator>::WriteText(string FileName) const
  {

    ofstream FileStream;
    FileStream.open(FileName.c_str());

#ifdef SELDON_CHECK_IO
    // Checks if the file was opened.
    if (!FileStream.is_open())
      throw IOError("Vector<Vect_Full>::WriteText(string FileName)",
		    "Unable to open file \"" + FileName + "\".");
#endif

    this->WriteText(FileStream);

    FileStream.close();

  }


  // Writes in a file.
  template <class T, class Allocator>
  void Vector<T, Vect_Full, Allocator>::WriteText(ofstream& FileStream) const
  {

#ifdef SELDON_CHECK_IO
    // Checks if the stream is ready.
    if (!FileStream.good())
      throw IOError("Vector<Vect_Full>::WriteText(ofstream& FileStream)",
                    "Stream is not ready.");
#endif

    if (this->GetLength()!=0)
      FileStream << (*this)(0);

    for (int i=1; i<this->GetLength(); i++)
      FileStream << "\t" << (*this)(i);

#ifdef SELDON_CHECK_IO
    // Checks if data was written.
    if (!FileStream.good())
      throw IOError("Vector<Vect_Full>::WriteText(ofstream& FileStream)",
                    string("Output operation failed. The output file may have been removed")
		    + "or there is no space left on device.");
#endif

  }


  // Reads a file.
  template <class T, class Allocator>
  void Vector<T, Vect_Full, Allocator>::Read(string FileName)
  {

    ifstream FileStream;
    FileStream.open(FileName.c_str());

#ifdef SELDON_CHECK_IO
    // Checks if the file was opened.
    if (!FileStream.open())
      throw IOError("Vector<Vect_Full>::Read(string FileName)",
		    "Unable to open file \"" + FileName + "\".");
#endif

    this->Read(FileStream);

    FileStream.close();

  }


  // Reads a file.
  template <class T, class Allocator>
  void Vector<T, Vect_Full, Allocator>::Read(ifstream& FileStream)
  {

#ifdef SELDON_CHECK_IO
    // Checks if the strem is ready.
    if (!FileStream.good())
      throw IOError("Vector<Vect_Full>::Read(ifstream& FileStream)",
                    "Stream is not ready.");
#endif

    int new_size;
    FileStream.read(reinterpret_cast<char*>(&new_size), sizeof(int));
    this->Reallocate(new_size);

    FileStream.read(reinterpret_cast<char*>(this->data_),
		    new_size * sizeof(value_type));

#ifdef SELDON_CHECK_IO
    // Checks if data was read.
    if (!FileStream.good())
      throw IOError("Vector<Vect_Full>::Read(ifstream& FileStream)",
                    string("Output operation failed.")
		    + " The intput file may have been removed or may not contain enough data.");
#endif    

  }


  // operator<< overloading for vectors.
  template <class T, class Storage, class Allocator>
  ostream& operator <<(ostream& out,
		       const Vector<T, Storage, Allocator>& V)
  {

    for (int i=0; i<V.GetLength()-1; i++)
      out << V(i) << '\t';
    if (V.GetLength()!=0)
      out << V(V.GetLength()-1);

    return out;

  }


} //namespace Seldon

#define FILE_VECTOR_CXX
#endif
